# frozen_string_literal: true

module MemoryProfiler

  class Monochrome

    def path(text)
      text
    end

    def string(text)
      text
    end

    def line(text)
      text
    end
  end

end
